import pandas as pd
import pathlib
import json



def check_struc(config):
    """
    Checks to ensure the config structure is valid
    """
    ## config requirements ##
    config_requiremnets = {
        'name' : {'type':str, 'nullable' : False},
        'description' : {'type':str, 'nullable' : False},
        'last_local_update' : {'type':str, 'nullable' : True},
        'download_url' : {'type':str, 'nullable' : False},
        'compression' : {'type':str, 'nullable' : True},
        'engine' : {'type':str, 'nullable' : True},
        'iter' : {'type':dict, 'nullable' : False},
        'assignments' : {'type':list, 'nullable' : False},
        'map' : {'type':dict, 'nullable' : False},
    }
    ## containers ##
    passing = True
    errors = []
    for k, v in config_requiremnets.items():
        if k not in config:
            ## check that config has the prop ##
            passing = False
            errors.append({
                'type' : 'config_structure',
                'error_msg' : '{0} is a required property, but was not found in the map config'.format(
                    k
                )
            })
        else:
            ## make sure its of the right type ##
            if config[k] is None and not v['nullable']:
                passing = False
                errors.append({
                    'type' : 'config_structure',
                    'error_msg' : '{0} is cannot be null'.format(
                        k
                    )
                })
            elif config[k] is not None:
                if not isinstance(config[k], v['type']):
                    errors.append({
                        'type' : 'config_structure',
                        'error_msg' : '{0} must be a {1}'.format(
                            k, v['type']
                        )
                    })
    ## return ##            
    return passing, errors

def check_map_type(map):
    """
    Iterates through map keys value pairs and ensures values are an allowed type
    """
    allowed_types = [
        'object', 'int32', 'int64', 'float32', 'float64',
        'boolean', 'interval', 'category', 'datetime64[ns, <tz>]',
        'bool'
    ]
    ## containers ##
    passing = True
    errors = []
    for k, v in map.items():
        if v not in allowed_types:
            passing = False
            errors.append({
                'error_type' : 'dtype',
                'error_msg' : '{0} type provided for {1} is not an accepted datatype'.format(
                    v, k 
                )
            })
    ## return ##
    return passing, errors

def check_data_folder():
    '''
    Runs on load and checks that the data folder exists. If it does not, it creates it
    '''
    ## get data ##
    data_dir = pathlib.Path(__file__).parent.parent.resolve() / 'Data'
    ## if it does not exist, create ##
    if not data_dir.exists():
        data_dir.mkdir()

def check_data():
    '''
    Runs on load and checks for the existence of local data for each map. If data does not exist,
    freshness meta data in the map is reset
    '''
    ## get maps ##
    maps_dir = pathlib.Path(__file__).parent.parent.resolve() / 'Maps'
    maps = list(maps_dir.glob('*.json'))
    ## check that maps were found ##
    if len(maps) == 0:
        return
    ## for each map, parse name, check for csv, and make updates as needed ##
    for map_path in maps:
        ## parse the name ##
        table = map_path.stem
        ## check for the csv ##
        csv_path = pathlib.Path(
            pathlib.Path(__file__).parent.parent.resolve() / 'Data' / '{0}.csv'.format(table)
        )
        csv_path.exists()
        if not csv_path.exists():
            ## if the csv does not exist, update the map file ##
            with open(map_path, 'r+') as config_file:
                data = json.load(config_file)
                data['last_local_update'] = None
                data['freshness']['last_freshness_check'] = None
                config_file.seek(0)  # reset file position to the beginning
                json.dump(data, config_file, indent=2)
                config_file.truncate()  # remove remaining part```



import pandas as pd
import pathlib
import time

from .nfelodcm import get_df

def test_all_maps():
    '''
    Attempts to load all maps
    '''
    print('Testing all available maps')
    map_loc = '{0}/Maps'.format(
        pathlib.Path(__file__).parent.resolve()
    )
    maps = []
    for f in pathlib.Path(map_loc).iterdir():
        if f.is_file() and f.suffix == ".json":
            maps.append(f.stem)
    print('     Found {0} maps'.format(len(maps)))
    if len(maps) > 0:
        for table_map in maps:
            print('     Loading {0}'.format(table_map))
            ## first load ##
            l1_start = time.time()
            get_df(table_map)
            l1_end = time.time()
            print('          1st load completed in {0} seconds'.format(round(
                l1_end-l1_start,3
            )))
            ## second load
            l2_start = time.time()
            get_df(table_map)
            l2_end = time.time()
            print('          2nd load completed in {0} seconds'.format(round(
                l2_end-l2_start,3
            )))





import json
import pathlib

class TableMap():
    """ load and update table maps """
    
    def __init__(self, table):
        self.table = table
        self.table_map_location = '{0}/Maps/{1}.json'.format(
            pathlib.Path(__file__).parent.parent.parent.resolve(),
            table
        )
        self.table_data_location = '{0}/Data/{1}.csv'.format(
            pathlib.Path(__file__).parent.parent.parent.resolve(),
            table
        )
        self.config = self.load_map()
        
    def load_map(self):
        """
        Loads the configuration json for a given table.
        """
        try:
            with open(self.table_map_location, 'r') as fp:
                config = json.load(fp)
                return config
        except Exception as e:
            print('ERROR: Could not load map for {0}'.format(self.table))
            print('       Use create_map from this class instance to create a new map')
            return None
    
    def update_map(self, new_config):
        """
        Loads the configuration json for a given table, and updates it with the new configuration
        """
        ## helper to update congfig -- wrapped in update_map as it should not be used ##
        ## outside of update_map, which also writes ##
        def update_config(old_config, new_config):
            for key, value in new_config.items():
                if key not in old_config:
                    print('WARN: {0} not in current config and will not be added'.format(key))
                else:
                    old_config[key] = value
            return old_config
        
        if self.config is None:
            print('WARN: No config exists for {0}. Use create_map() to create a table map'.format(self.table))
            return
        else:
            ## update ##
            config = update_config(self.config, new_config)
            ## write ##
            with open(self.table_map_location, 'w') as fp:
                json.dump(config, fp, indent=2)
            ## reload ##
            self.config = self.load_map()

    def create_map(self, new_config):
        """
        Creates a new table map from the given configuration
        """
        ## make sure new_confi is properly formatted json ##
        try:
            json.dumps(new_config)
        except Exception as e:
            print('ERROR: Config passed was not json. No mapping was created'.format(e))
            return
        ## write ##
        with open(self.table_map_location, 'w') as fp:
            json.dump(new_config, fp, indent=2)
        ## reload ##
        self.config = self.load_map()



import pandas as pd
import numpy
import pathlib
import json
import datetime

## script that runs on package import to set season states ##
##      last partial week is the last week with any played game ##
##      last full week is the last week with no unplayed games ##
##      next week is the first week with no played games ##
## This script leverages nflgamedata to determine states ##

def set_season_state():
    ## Try to set the state with data, but as a backup, infer from month ##
    try:
        games = pd.read_csv(
            "https://raw.githubusercontent.com/nflverse/nfldata/master/data/games.csv"
        )
        ## demarcate played games ##
        games['played'] = numpy.where(
            ~pd.isnull(games['result']),
            1,
            0
        )
        
        ## get weekly pct played ##
        weeks = games.groupby(['season', 'week']).agg(
            pct_played = ('played', 'mean')
        ).reset_index()
        weeks = weeks.sort_values(
            by=['season', 'week'],
            ascending=[True, True]
        ).reset_index(drop=True)
        
        ## cut data by pct played ##
        partial = weeks[weeks['pct_played'] > 0].copy()
        full = weeks[weeks['pct_played'] == 1].copy()
        next = weeks[weeks['pct_played'] == 0].copy()
        
        ## open global variables json and write values ##
        global_variables = None
        with open('{0}/global_variables.json'.format(pathlib.Path(__file__).parent.resolve()), 'r') as fp:
            ## load config ##
            global_variables = json.load(fp)
        
        ## set values ##
        global_variables['season_states']['last_full_week']['week'] = int(full.iloc[-1]['week'])
        global_variables['season_states']['last_full_week']['season'] = int(partial.iloc[-1]['season'])
        global_variables['season_states']['last_partial_week']['week'] = int(partial.iloc[-1]['week'])
        global_variables['season_states']['last_partial_week']['season'] = int(partial.iloc[-1]['season'])
        try:
            global_variables['season_states']['next_week']['week'] = int(next.iloc[0]['week'])
            global_variables['season_states']['next_week']['season'] = int(next.iloc[0]['season'])
        except:
            global_variables['season_states']['next_week']['week'] = numpy.nan
            global_variables['season_states']['next_week']['season'] = numpy.nan
        ## write ##
        with open('{0}/global_variables.json'.format(pathlib.Path(__file__).parent.resolve()), 'w') as fp:
            json.dump(global_variables, fp, indent=4)
        
    except Exception as e:
        print('WARN: Could not set season states with nflgamedata. Inferring from date instead.')
        print('       {0}'.format(e))
        today = datetime.datetime.today()
        
        ## open global variables json and write values ##
        global_variables = None
        with open('{0}/global_variables.json'.format(pathlib.Path(__file__).parent.resolve), 'r') as fp:
            ## load config ##
            global_variables = json.load(fp)
        
        ## set values ##
        global_variables['season_states']['last_full_week']['week'] = numpy.nan
        global_variables['season_states']['last_partial_week']['week'] = numpy.nan
        global_variables['season_states']['last_partial_week']['season'] = numpy.nan
        global_variables['season_states']['next_week']['week'] = numpy.nan
        global_variables['season_states']['next_week']['season'] = numpy.nan
        if today.month < 9:
            global_variables['season_states']['last_full_week']['season'] = today.year - 1
        else:
            global_variables['season_states']['last_full_week']['season'] = today.year



import pandas as pd
import numpy

import nfelodcm.nfelodcm.Utilities as utils
import nfelodcm.nfelodcm.Engine.Assignments as assignments

class DataPull():
    """
    Downloads CSVs provided a map that contains url, fields, and iter details
    """
    
    def __init__(self, config):
        self.config = config
        self.columns, self.dtypes = utils.extract_map(config['map'])
        self.pulled_data = []
        self.pulled_df = None
    
    def pull_data(self, url):
        """
        Pulls data from the url and returns a pandas dataframe
        """
        ## logic for retries ##
        retries_remaining = 3
        pull_error = None
        while retries_remaining > 0:
            try:
                ## pull data ##
                df = pd.read_csv(
                    url,
                    usecols=self.columns,
                    dtype=self.dtypes,
                    engine=self.config['engine']
                )
                ## add pulled data to container ##
                self.pulled_data.append(df)
                ## end if successful ##
                break
            except Exception as e:
                print('ERROR: Failed to pull data from {0}. Retrying...'.format(
                    url
                ))
                pull_error = e
                retries_remaining -= 1
        ## if retries are exhausted, raise an error ##
        if retries_remaining == 0:
            raise ValueError(
                'ERROR: Failed to pull data from {0} after 3 retries: {1}'.format(
                    url,
                    pull_error
                )
            )
        
    def handle_pull(self):
        """
        Handles the pull based on the iter type. If the iter is a loop, it will form download urls and loop,
        otherwise it will pull simply pull the download url
        """
        ## handle pull ##
        if self.config['iter']['type'] is None:
            ## if not iter type, assume single download url ##
            self.pull_data(self.config['download_url'])
        elif self.config['iter']['type'] == 'season':
            ## if season iter, loop through seasons ##
            for season in range(
                self.config['iter']['start'],
                utils.current_season() + 1
            ):
                ## form download url ##
                url = self.config['download_url'].format(season)
                ## pull data ##
                self.pull_data(url)
        else:
            raise ValueError('ERROR: Iter type not recognized.')
    
    def concat_data(self):
        """
        Once data is pulled, it will handle any necessary concatenation
        """
        if len(self.pulled_data) == 0:
            raise ValueError('ERROR: A data request was made, but no data was pulled.')
        elif len(self.pulled_data) == 1:
            self.pulled_df = self.pulled_data[0]
        else:
            df = pd.concat(self.pulled_data)
            self.pulled_df = df.reset_index(drop=True)
    
    def apply_assignments(self):
        """
        Applies any assignments to the pulled data
        """
        ## apply assignments ##
        if len(self.config['assignments']) > 0:
            for assignment in self.config['assignments']:
                ## apply assignment ##
                self.pulled_df = assignments.assign(
                    self.pulled_df,
                    assignment
                )
    
    def update_data(self):
        """
        Wrrapper to handle the pull if called
        """
        ## handle pull ##
        self.handle_pull()
        ## concat data ##
        self.concat_data()
        ## apply assignments ##
        self.apply_assignments()



from .nfelodcm.nfelodcm import get_df
from .nfelodcm.nfelodcm import load
from .nfelodcm.nfelodcm import get_map
from .nfelodcm.nfelodcm import get_season_state
from .nfelodcm.nfelodcm import list_tables
from .nfelodcm.tests import test_all_maps


import pandas as pd
import numpy

def temp_players_handle(df):
    """
    Renames the the players.csv columns (draftround) to match
    the old naming. This is a temporary fix as nflfastR is updating
    pipelines and does not plan to address this bug

    When the new pipelines are officially released, players will point
    to the new file
    """
    ## rename ##
    df = df.rename(columns={'draftround': 'draft_round'})
    return df


import pandas as pd

qb_name_repl = {
    "Robert Griffin": "Robert Griffin III",
    "A.J. McCarron": "AJ McCarron",
    "Mitch Trubisky": "Mitchell Trubisky"
}

def fix_elo_qb_names(df):
    '''
    Standardizes QB names to match fastr
    '''
    ## fix player names ##
    df['qb1'] = df['qb1'].replace(qb_name_repl)
    df['qb2'] = df['qb2'].replace(qb_name_repl)
    ## return ##
    return df


from .extract_map import extract_map
from .retrieve_season_state import current_season
from .set_season_state import set_season_state
from .checks import check_struc
from .checks import check_map_type
from .checks import check_data
from .checks import check_data_folder
from .load_maps import load_maps


from .fastr_team_repl import team_id_repl as fastr_team_id_repl
from .penalty_formatting import penalty_formatting
from .desc_based_plays import desc_based_plays
from .score_clean import score_clean
from .format_qb_file import fix_elo_qb_names
from .format_player_file import fix_fastr_qb_names
from .temp_players_handle import temp_players_handle


import pandas as pd

## assignment functions that replaces fastr team names with legacy team names ##
repl = {
    'LV' : 'OAK',
    'SD' : 'LAC',
    'STL' : 'LAR',
    'LA' : 'LAR',
}

def team_id_repl(df):
    """
    Replaces fastr team ids with a legacy nfelo ids.
    """
    ## if a col with team names exists, replace it ##
    for col in [
        'home_team', 'away_team', 'team_abbr',
        'posteam', 'defteam', 'penalty_team',
        'side_of_field', 'timeout_team', 'td_team',
        'return_team', 'possession_team',
        'recent_team', 'opponent_team',
        'draft_team', 'draft_club'
    ]:
        if col in df.columns:
            df[col] = df[col].replace(repl)
    ## if game if is also in the DF, replace it
    if 'game_id' in df.columns:
        ## id col ##
        df['game_id'] = (
            df['season'].astype('str') + '_' +
            df['week'].astype('str').str.zfill(2) + '_' +
            df['away_team'] + '_' +
            df['home_team']
        )
    return df



import pandas as pd

qb_name_repl = {
    "Jacob Dolegala": "Jake Dolegala",
    "Robert Griffin": "Robert Griffin III",
    "Thad Lewis": "Thaddeus Lewis",
    "E.J. Manuel": "EJ Manuel",
    "A.J. McCarron": "AJ McCarron",
    "Matthew McGloin": "Matt McGloin",
    "Vincent Testaverde": "Vinny Testaverde",
    "Mike Vick": "Michael Vick",
    "Phillip Walker": "P.J. Walker"
}


def fix_fastr_qb_names(df):
    '''
    Standardizes QB names to match elo
    '''
    ## fix player names ##
    df['full_name'] = df['full_name'].replace(qb_name_repl)
    ## return ##
    return df


import pandas as pd
import numpy

manual_clean_dict = {
    "2009_17_IND_BUF": {
        "home_score": 30,
        "away_score": 7
    },
    "2013_07_CIN_DET": {
        "home_score": 24,
        "away_score": 27
    },
    "2015_06_ARI_PIT": {
        "home_score": 25,
        "away_score": 13
    },
    "2015_09_PHI_DAL": {
        "home_score": 27,
        "away_score": 33
    },
    "2015_15_KC_BAL": {
        "home_score": 14,
        "away_score": 34
    },
    "2016_01_MIN_TEN": {
        "home_score": 16,
        "away_score": 25
    },
    "2016_05_NE_CLE": {
        "home_score": 13,
        "away_score": 33
    }
}

def score_clean(df):
    '''
    Manually clean scores that are incorrect in the
    game file
    '''
    ## create a home and away map ##
    home_map = {k:manual_clean_dict[k]['home_score'] for k in manual_clean_dict.keys()}
    away_map = {k:manual_clean_dict[k]['away_score'] for k in manual_clean_dict.keys()}
    ## apply ##
    df['home_score'] = df['game_id'].map(home_map).combine_first(df['home_score'])
    df['away_score'] = df['game_id'].map(away_map).combine_first(df['away_score'])
    ## return 
    return df